#-------------------------------------------------------------------------------
# Name:        mols
# Purpose:     Mol class to operate with molecules
#
# Author:      Pavel Polishchuk
#
# Created:     11.01.2013
# Copyright:   (c) Pavel Polishchuk 2013-2015
# Licence:     BSD 3-clause
#-------------------------------------------------------------------------------

from itertools import combinations
import copy


class Mol3:
    # double bonds can be:
    # 2 - non-steric or undefined double bond (2, 0)
    # 21 - Z double bond (2, 1)
    # 22 - E double bond (2, 2)
    # 23 - cyclic double bond in rings of size up to 7 (such bonds are always cis) (2, 3)

    def __init__(self):
        self.atoms = {}
        self.bonds = dict()
        self.title = ""
        self.stereo = False

    def AddAtom(self, id, label, x, y, z, formal_charge):
        self.atoms[id] = {'label': label, 'x': x, 'y': y, 'z': z,
                          'property': {},    # will contain dicts of type 'elm': {'label': ['C'], 'value': 'C'}
                          'formal_charge': formal_charge}
        self.bonds[id] = dict()

    def AddBond(self, id1, id2, bond_type):
        if id1 not in self.bonds.keys():
            self.bonds[id1] = dict()
        if id2 not in self.bonds.keys():
            self.bonds[id2] = dict()
        # bond value is a tuple: 1 - bond order, 2 - double bond stereo type (0 - unspecified, 1 - cis, 2 -trans, 3 - cyclic)
        self.bonds[id1][id2] = self.bonds[id2][id1] = (bond_type, 0)

    def GetBondOrder(self, id1, id2):
        return self.bonds.get(id1, dict()).get(id2, (0, 0))[0]

    def GetBondType(self, id1, id2):
        bond_order = self.GetBondOrder(id1, id2)
        if self.stereo and bond_order == 2 and self.bonds[id1][id2][1] != 0:
            return self.bonds[id1][id2][0] * 10 + self.bonds[id1][id2][1]
        else:
            return bond_order

    def SetDoubleBondConfig(self, id1, id2, bond_stereo):
        if bond_stereo not in [0, 1, 2, 3]:
            raise Exception('Wrong double bond stereo!')
        self.bonds[id1][id2] = self.bonds[id2][id1] = (2, bond_stereo)

    def _Path(self, start_atom, list_atom, cycles_local, visited, size_range):
        for a in self.bonds[start_atom].keys():
            if len(list_atom) <= max(size_range) and a not in list_atom and a not in visited:
                self._Path(a, list_atom + [a], cycles_local, visited, size_range)
            elif len(list_atom) in size_range and a == list_atom[0]:
                if tuple(set(sorted(list_atom))) not in cycles_local:
                    cycles_local.add(tuple(set(sorted(list_atom))))

    def GetCycles(self, min_size, max_size):
        cycles = set()
        visited = set()  # atoms which already has been tested as a cycle member can be excluded from further cycles checks
        for a in sorted(self.atoms.keys()):
            visited.add(a)
            if len(self.bonds[a].keys()) > 1:
                self._Path(a, [a], cycles, visited, range(min_size, max_size + 1))
        return cycles

    def SetCyclicDoubleBondsCis(self, min_size=3, max_size=7):
        cycles = self.GetCycles(min_size, max_size)
        for cycle in cycles:
            for a1, a2 in combinations(cycle, 2):
                if self.GetBondOrder(a1, a2) == 2:
                    self.bonds[a1][a2] = self.bonds[a2][a1] = (2, 3)


class Mol4(Mol3):

    def __GetNumberConnectedComponents(self, atoms):

        res = 0
        visited = [False] * len(atoms)

        def dfs(i):
            visited[i] = True
            for a in self.bonds[atoms[i]].keys():
                if a in atoms and not visited[atoms.index(a)]:
                    dfs(atoms.index(a))

        for i in range(len(atoms)):
            if not visited[i]:
                dfs(i)
                res += 1

        return res

    def __GenConnectedSubgraphs(self, not_visited, min_num_atoms=4, max_num_atoms=4, curr_subset=set(), neighbors=set(), res=[]):

        if min_num_atoms <= len(curr_subset) <= max_num_atoms:
            res.append(curr_subset)

        if not curr_subset:
            candidates = set(not_visited)
        else:
            candidates = not_visited.intersection(neighbors)

        if candidates and len(curr_subset) < max_num_atoms:
            for a in candidates:
                not_visited.remove(a)
                tmp1 = set(not_visited)
                tmp2 = set(curr_subset)
                tmp2.add(a)
                tmp3 = not_visited.intersection(self.bonds[a].keys())
                tmp3 = neighbors.union(tmp3)
                self.__GenConnectedSubgraphs(tmp1, min_num_atoms=min_num_atoms, max_num_atoms=max_num_atoms, curr_subset=tmp2, neighbors=tmp3, res=res)

    def __GetAllNeighbours(self, atoms):
        output = set(atoms)
        for a in atoms:
            output = output.union(self.bonds[a].keys())
        return output


    def GetAtomsCombinations(self, min_num_components=1, max_num_components=2, min_num_atoms=4, max_num_atoms=4, noH=False):

        def CheckIntersection(sets, neighbour_sets, ids):
            if set.intersection(*[sets[i] for i in ids]):
                return True
            for i in ids:
                for j in ids:
                    if i != j:
                        if sets[i].intersection(neighbour_sets[j]):
                            return True
            return False

        if noH:
            atoms = set(a for a in self.atoms.keys() if self.atoms[a]["label"] != 'H')
        else:
            atoms = set(self.atoms.keys())

        # storage of results
        res = []
        # if only single component fragments are looking for then there is no need to search for fragments smaller than nim_num_atoms
        if max_num_components == 1:
            self.__GenConnectedSubgraphs(atoms, min_num_atoms=min_num_atoms, max_num_atoms=max_num_atoms, res=res)
        else:
            self.__GenConnectedSubgraphs(atoms, min_num_atoms=1, max_num_atoms=max_num_atoms, res=res)

        # get neighbours
        nb = [self.__GetAllNeighbours(v) for v in res]

        results = []

        for n in range(min_num_components, max_num_components + 1):
            for comb in combinations(range(len(res)), n):
                # if min_num_atoms <= sum(len(res[i]) for i in comb) <= max_num_atoms and (len(comb) == 1 or not set.intersection(*[nb[i] for i in comb])):
                if min_num_atoms <= sum(len(res[i]) for i in comb) <= max_num_atoms and (len(comb) == 1 or not CheckIntersection(res, nb, comb)):
                    results.append(tuple(set.union(*[res[i] for i in comb])))

        return results


class SmilesMol3(Mol4):

    tosmileskeys = {0: '.', 1: '-', 2: '=', 3: '#', 4: ':', 8: '~'}

    def __tosmiles(self, bond):
        return self.tosmileskeys[bond]

    def __getSmiles(self, trace, inter, labels_dict):
        # trace: atom ids
        # inter: selected atom id
        strace = set(trace)
        iterlist = set(self.bonds[inter].keys()).intersection(self.sub).difference(trace[-2:])

        # get atom label
        smi = [labels_dict[inter]]

        self.nextnumb = 1

        concat = []
        stoplist = []
        iterlen = len(iterlist) - 1
        for b, i in enumerate(sorted(list(iterlist), key=self.levels.get)):
            if i in strace:
                if i not in stoplist:
                    # костыль для циклов. чтоб не было 2х проходов.
                    cyc = self.nextnumb
                    self.nextnumb += 1
                    concat += [(i, cyc, inter)]
                    smi[0] += '%s%d' % (self.__tosmiles(self.GetBondOrder(inter, i)), cyc)
                if b == iterlen and len(smi) > 3:
                    smi[-1] = smi[-3] = ''
                continue
            deep = self.__getSmiles(copy.copy(trace + [i]), i, labels_dict)
            strace.update(deep[0])

            for j in deep[2]:
                if j[0] == inter:
                    stoplist += [j[2]]
                    smi[0] += '%s%d' % (self.__tosmiles(self.GetBondOrder(inter, j[2])), j[1])
                else:
                    concat.append(j)

            smi += ['(' if iterlen - b else '', '%s' % self.__tosmiles(self.GetBondOrder(inter, i)) + deep[1],
                    ')' if iterlen - b else '']

        return strace, ''.join(smi), concat

    def __get_feature_signatures(self, ids, labels):
        feature_signatures = []
        for i, label_i in zip(ids, labels):
            sign = []
            for j, label_j in zip(ids, labels):
                if i != j:
                    bond_order = self.GetBondOrder(i, j)
                    if bond_order > 0:
                        sign.append((label_j, bond_order))
            feature_signatures.append((label_i,) + tuple(sorted(sign)))
        return tuple(feature_signatures)

    def __getRanks(self, sub, labels):
        prev_len = 0
        signatures = self.__get_feature_signatures(sub, labels)
        while len(sub) > len(set(signatures)) > prev_len:
            prev_len = len(set(signatures))
            signatures = self.__get_feature_signatures(sub, signatures)
        s = sorted(signatures)
        return {atom: s.index(sign) for atom, sign in zip(sub, signatures)}

    def get_name(self, sub, labels):

        # if {10, 11, 12}.issubset(sub):
        #     # if 'H' in labels:
        #     if set(sub).intersection([24, 20]):
        #         print(sub)

        # self.nextnumb = self.__numb()
        self.sub = set(sub)
        # self.levels = self.__getWeininger(sub, labels)
        self.levels = self.__getRanks(sub, labels)
        inter = min(self.levels, key=self.levels.get)

        res = [self.__getSmiles([inter], inter, dict(zip(sub, labels)))]
        visited = res[0][0]
        while visited != self.sub:
            remained = {k: self.levels[k] for k in self.sub.difference(visited)}
            inter = min(remained, key=remained.get)
            res.append(self.__getSmiles([inter], inter, dict(zip(sub, labels))))
            visited = visited.union(res[-1][0])

        # to get canonical multi-component SMILES
        res = sorted([r[1] for r in res])

        return '.'.join(res)

    # def __numb(self):
    #     i = 1
    #     while True:
    #         yield i
    #         i += 1




__author__ = 'pavel'
# Basic functions to work with descriptors names


def get_smile(sirms_string):
    return sirms_string.rsplit('|', 1)[1]


def get_atomcount(sirms_string):
    return int(sirms_string.split('|', 6)[5])


def get_atom_labeling(sirms_string):
    return sirms_string.rsplit('|', 2)[1]


def get_mix_single(sirms_string):
    return sirms_string.split('|', 2)[1]


def invert_num_prob_type(sirms_string):
    tmp = sirms_string.split('|')
    if tmp[2] == 'n':
        tmp[2] = 'p'
    else:
        tmp[2] = 'n'
    return '|'.join(tmp)


def insert_reaction_info(sirms_string, prod_react):
    """
    prod_react: p or r or pr
    """
    return prod_react + '|' + sirms_string.split('|', 1)[1]


def split_by_reaction_part(sirms_name):
    return sirms_name.split('|', 1)


def join_reaction_part(react_part, common_part):
    return react_part + '|' + common_part


def create_full_name(prod_react, single_mix, num_prob, atom_count, atom_labeling, smiles):
    return prod_react + '|' + single_mix + '|' + num_prob + '|||' + str(atom_count) + '|||' + \
           atom_labeling + '|' + smiles





#-------------------------------------------------------------------------------
# Name:        canon
# Purpose:     generate canonic sirms names and stereodescriptors
#
# Author:      Pavel Polishchuk
#
# Created:     11.01.2013
# Copyright:   (c) Pavel Polishchuk 2013
# Licence:     BSD 3-clause
#-------------------------------------------------------------------------------

from itertools import combinations, permutations, chain, product
from collections import defaultdict
from math import acos, sqrt, degrees
from .ppgfunctions import SortTwoLists
import os
import json

# Global constants

dumb_label_set = ['A', 'B', 'C', 'D']

elements = ['H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na', 'Mg',
            'Al', 'Si', 'P', 'S', 'Cl', 'Ar', 'K', 'Ca', 'Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe',
            'Co', 'Ni', 'Cu', 'Zn', 'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr', 'Rb', 'Sr', 'Y',
            'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd', 'In', 'Sn', 'Sb', 'Te',
            'I', 'Xe', 'Cs', 'Ba', 'La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb',
            'Dy', 'Ho', 'Er', 'Tm', 'Yb', 'Lu', 'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt',
            'Au', 'Hg', 'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac', 'Th', 'Pa',
            'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk', 'Cf', 'Es', 'Fm', 'Md', 'No', 'Lr', 'Rf',
            'Db', 'Sg', 'Bh', 'Hs', 'Mt', 'Ds', 'Rg', 'Uub', 'Uut', 'Uuq', 'Uup', 'Uuh',
            'Uus', 'Uuo']

#base simplex types represented as sets of vertex neighbors
s_types = {(0, 0, 0, 0): 1, (0, 0, 1, 1): 2, (1, 1, 1, 1): 3, (0, 1, 1, 2): 4, (1, 1, 1, 3): 5,
           (1, 1, 2, 2): 6, (0, 2, 2, 2): 7, (1, 2, 2, 3): 8, (2, 2, 2, 2): 9, (2, 2, 3, 3): 10,
           (3, 3, 3, 3): 11}

bin_s_types = {(1, 0, 0, 1, 1, 1): 8, (0, 1, 0, 1, 1, 0): 6, (0, 1, 1, 1, 0, 1): 8, (1, 1, 0, 0, 1, 0): 6,
               (0, 0, 1, 0, 1, 0): 4, (0, 1, 1, 0, 0, 0): 4, (1, 0, 0, 0, 0, 0): 2, (0, 1, 1, 1, 1, 1): 10,
               (1, 0, 0, 1, 0, 0): 4, (1, 1, 1, 1, 1, 0): 10, (0, 0, 1, 0, 0, 1): 4, (0, 1, 1, 0, 1, 1): 8,
               (0, 0, 1, 1, 0, 0): 3, (0, 1, 1, 1, 0, 0): 6, (1, 0, 1, 1, 0, 1): 9, (0, 0, 1, 0, 1, 1): 5,
               (1, 1, 1, 0, 1, 0): 8, (0, 1, 1, 0, 0, 1): 7, (1, 0, 1, 0, 0, 0): 4, (0, 0, 1, 1, 1, 1): 8,
               (1, 1, 1, 1, 0, 0): 8, (1, 0, 1, 1, 1, 1): 10, (1, 0, 0, 1, 1, 0): 5, (0, 0, 1, 0, 0, 0): 2,
               (1, 1, 1, 0, 0, 1): 8, (0, 1, 1, 0, 1, 0): 6, (1, 0, 1, 0, 1, 1): 8, (0, 0, 1, 1, 0, 1): 6,
               (0, 0, 0, 1, 1, 0): 4, (1, 0, 1, 1, 0, 0): 6, (1, 1, 1, 0, 1, 1): 10, (1, 0, 1, 0, 0, 1): 6,
               (1, 1, 1, 1, 0, 1): 10, (0, 0, 1, 1, 1, 0): 6, (0, 0, 0, 1, 0, 1): 4, (0, 0, 0, 0, 0, 0): 1,
               (1, 1, 1, 0, 0, 0): 5, (1, 0, 1, 0, 1, 0): 7, (0, 1, 0, 0, 1, 0): 3, (0, 0, 0, 1, 1, 1): 7,
               (1, 1, 0, 1, 1, 0): 8, (0, 0, 0, 0, 1, 1): 4, (1, 0, 0, 1, 0, 1): 6, (0, 1, 0, 0, 0, 1): 4,
               (0, 0, 0, 1, 0, 0): 2, (1, 1, 0, 1, 0, 1): 8, (0, 1, 0, 1, 0, 0): 4, (0, 0, 0, 0, 0, 1): 2,
               (1, 1, 0, 0, 0, 0): 4, (0, 1, 0, 0, 1, 1): 6, (1, 0, 0, 0, 1, 0): 4, (1, 1, 0, 1, 1, 1): 10,
               (0, 1, 0, 1, 1, 1): 8, (0, 0, 0, 0, 1, 0): 2, (1, 1, 0, 0, 1, 1): 9, (0, 1, 0, 0, 0, 0): 2,
               (1, 0, 0, 0, 0, 1): 3, (1, 1, 0, 1, 0, 0): 7, (0, 1, 0, 1, 0, 1): 5, (0, 1, 1, 1, 1, 0): 9,
               (1, 1, 0, 0, 0, 1): 6, (1, 0, 1, 1, 1, 0): 8, (1, 1, 1, 1, 1, 1): 11, (1, 0, 0, 0, 1, 1): 6}


def LoadSirmsDict():
    f = open(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'short_sirms_dict.json'))
    tmp = json.load(f)
    f.close()
    return (tmp)


##    f = open(os.path.join(os.path.dirname(sys.argv[0]), 'short_sirms_dict.txt'), 'tr')
##    sirms_dict = {}
##    for line in f:
##        tmp = line.strip().split('\t')
##        sirms_dict[tmp[0]] = tmp[1]
##    f.close()
##    return(sirms_dict)

def det(a):
    return (a[0][0] * (a[1][1] * a[2][2] - a[2][1] * a[1][2])
            - a[1][0] * (a[0][1] * a[2][2] - a[2][1] * a[0][2])
            + a[2][0] * (a[0][1] * a[1][2] - a[1][1] * a[0][2]))


def GetStereoRL(atom_elements, coord):
    atom_numbers = [elements.index(el) for el in atom_elements]
    # sort by atomic number, highest atom is last, lowest atom is first
    atom_numbers, coord = SortTwoLists(atom_numbers, coord)
    # calc diffefence between coord of the lowest atom and all other atoms
    b = [[j - coord[0][i] for i, j in enumerate(elm)] for elm in coord[1:]]
    # calc the signed volume of parallelepiped
    d = det(b)
    if d > 0: return ('R')
    if d < 0: return ('L')
    return (None)


def GetNormalCoord(c):
    # calc normal vector to the plane represented by three points (list of lists)
    # http://wiki.mirgames.ru/%D0%BD%D0%BE%D1%80%D0%BC%D0%B0%D0%BB%D1%8C#naxozhdenie_vektora_normali_k_ploskosti_postroennoj_po_trem_tochkam
    A = c[0][1] * (c[1][2] - c[2][2]) + c[1][1] * (c[2][2] - c[0][2]) + c[2][1] * (c[0][2] - c[1][2])
    B = c[0][2] * (c[1][0] - c[2][0]) + c[1][2] * (c[2][0] - c[0][0]) + c[2][2] * (c[0][0] - c[1][0])
    C = c[0][0] * (c[1][1] - c[2][1]) + c[1][0] * (c[2][1] - c[0][1]) + c[2][0] * (c[0][1] - c[1][1])
    return (A, B, C)


def GetDihedralAngle(coord1, coord2):
    # return dihedral angle in dergees
    A1, B1, C1 = GetNormalCoord(coord1)
    A2, B2, C2 = GetNormalCoord(coord2)
    return (degrees(
        acos((A1 * A2 + B1 * B2 + C1 * C2) / sqrt(A1 ** 2 + B1 ** 2 + C1 ** 2) / sqrt(A2 ** 2 + B2 ** 2 + C2 ** 2))))


def GetStereoZE(coord):
    # coord - coordinates of 4 consequentially bonded atoms
    # http://www.mathsisfun.com/geometry/dihedral-angles.html
    dihedral_angle = GetDihedralAngle(coord[0:3], coord[1:4])
    if 175 <= dihedral_angle <= 180:
        return ("E")
    if 0 <= dihedral_angle <= 5:
        return ("Z")
    return (None)


def GetStereoP(coord):
    # checks whether simplex is planar
    # coord - sorted coordinates of vertexes in order [1,1,2,4] (vertex degree)
    dihedral_angle = GetDihedralAngle(coord[0:2] + [coord[3]], coord[1:4])
    if (0 <= dihedral_angle <= 5) or (175 <= dihedral_angle <= 180):
        return (True)
    else:
        return (False)


def GetCanonNameByDict(labels, bonds, sirms_dict):
    uniq_labels = sorted(set(labels))
    dumb_labels = [dumb_label_set[uniq_labels.index(a)] for a in labels]
    s_name = ','.join(dumb_labels) + '|' + ','.join(map(str, bonds))
    canon_bonds = sirms_dict[s_name]
    return (','.join(sorted(labels)) + '|' + canon_bonds)


def GetSirmsType(bonds):
    ##    return(bin_s_types[tuple(b > 0 for b in bonds)])
    s_vertexes = [0, 0, 0, 0]
    for i, b in enumerate(combinations([0, 1, 2, 3], 2)):
        if bonds[i] > 0:
            s_vertexes[b[0]] += 1
            s_vertexes[b[1]] += 1
    return (s_types[tuple(sorted(s_vertexes))])


def GetSirmsType2(mol, atoms):
    b = []
    # atoms = set(atoms)
    for a in atoms:
        # b.append(len(set(mol.bonds[a].keys()).intersection(atoms)))
        tmp = 0
        for a1 in mol.bonds[a].keys():
            if a1 in atoms:
                tmp += 1
        b.append(tmp)
    idx = tuple(sorted(b))
    return (s_types[idx])


def GenCanonName(labels, bonds, a):
    # form simplex as dict
    s = {a_id: {"label": label} for label, a_id in zip(labels, a)}
    for i, b in enumerate(combinations(a, 2)):
        s[b[0]][b[1]] = s[b[1]][b[0]] = bonds[i]
    # get simplex type
    s_type = GetSirmsType(bonds)
    # create dict with labels and corresponding ids
    d = defaultdict(list)
    for k, v in s.items():
        d[v["label"]].append(k)
    # generate all combinations of ids of each label and combine generated sequences in specific order
    max_bond_value = [-1, -1, -1, -1, -1, -1]  # list which will store the maximum found value of bonds
    p = {k: permutations(v) for k, v in d.items()}
    for k in product(*map(p.get, sorted(p))):
        x = [s[i[0]][i[1]] for i in combinations(list(chain.from_iterable(k)), 2)]
        if x > max_bond_value:
            max_bond_value = x
    # return canonic name
    return (','.join(sorted([v["label"] for v in s.values()])) + '|' + ','.join(map(str, max_bond_value))) + '|' + str(
        s_type)



#-------------------------------------------------------------------------------
# Name:        sdf
# Purpose:     operations with sdf files (reading)
#
# Author:      Pavel Polishchuk
#
# Created:     11.01.2013
# Copyright:   (c) Pavel Polishchuk 2013-2015
# Licence:     BSD 3-clause
#-------------------------------------------------------------------------------


import re
from collections import OrderedDict

from .mols import SmilesMol3 as Mol
from .files import ReadPropertyRange, RangedLetter


formal_charges_table = {'0': 0,
                        '1': 3,
                        '2': 2,
                        '3': 1,
                        '4': 0,
                        '5': -1,
                        '6': -2,
                        '7': -3}


def molstr_to_Mol(molstr):

    mol = Mol()

    mol.title = molstr[0]

    natoms = int(molstr[3][0:3])
    nbonds = int(molstr[3][3:6])

    # read atoms
    id = 0
    for line in molstr[4:4 + natoms]:
        x = float(line[0:10])
        y = float(line[10:20])
        z = float(line[20:30])
        label = line[30:33].strip()
        formal_charge = line[36:39].strip()
        id += 1
        mol.AddAtom(id, label, x, y, z, formal_charges_table.get(formal_charge, 0))

    # read bonds
    for line in molstr[4 + natoms:4 + natoms + nbonds]:
        id1 = int(line[0:3])
        id2 = int(line[3:6])
        bond_type = int(line[6:9])
        mol.AddBond(id1, id2, bond_type)

    return mol


def add_property_to_atoms(mol, data_dict, fsetup):
    for prop_name in data_dict.keys():
        prop_range = ReadPropertyRange(fsetup, prop_name)
        for i, a in enumerate(sorted(mol.atoms.keys())):
            # if value is numeric use ranges, if value is string use it as label itself
            label = data_dict[prop_name][i] if type(data_dict[prop_name][i]) is str else RangedLetter(data_dict[prop_name][i], prop_range)
            # if label is represented as A|D three labels should be created A, D and AD
            label_full = label.split('|')
            if len(label_full) > 1:
                label_full.insert(0, label.replace('|', ''))
            mol.atoms[a]['property'][prop_name] = {'value': data_dict[prop_name][i],
                                                   'label': label_full}


def get_sdf_field(sdf_lines, field_name):

    # case-sensitive field name search

    i = 0

    patt = "^> +<%s>" % field_name

    # find field
    for i, line in enumerate(sdf_lines):
        if re.match(patt, line):
            break

    if i == len(sdf_lines) - 1:
        print('Field %s was not found in the input file.' % field_name)
        return None

    # retrieve data if field was found
    i += 1
    data = []
    while i < len(sdf_lines):
        if sdf_lines[i].find('>  <') == 0 or sdf_lines[i].find('> <') == 0 or sdf_lines[i].rstrip() == '$$$$':
            break
        if sdf_lines[i]:
            data.append(sdf_lines[i])
        i += 1

    return data


def ReadSDF(fname, id_field_name, opt_diff, fsetup, parse_stereo=False):
    """
    INPUT: sdf-filename
    OUTPUT: dict of molecules, where key is the title of the moleculae taken from the first line of mol-record
    """

    def _MolstrToMol(molstr, id_field_name, opt_diff, parse_stereo, num_mol):

        mol = molstr_to_Mol(molstr)

        # get mol name from sdf field (multi-line data fields are concatenated via dot)
        if id_field_name is not None:
            mol.title = '.'.join(get_sdf_field(molstr, id_field_name))
        if not mol.title:
            mol.title = 'auto_generated_id_' + str(num_mol)
        mol.stereo = parse_stereo

        # read properties from sdf fields
        data_dict = dict()
        for diff in opt_diff:
            data = get_sdf_field(molstr, diff)
            if data is not None and len(data) == 1:
                s = data[0].split(';')
                try:
                    data_dict[diff] = [float(el.replace(",", ".")) if el != "" else None for el in s]
                except ValueError:
                    data_dict[diff] = [el if el != "" else None for el in s]

        # add labels
        add_property_to_atoms(mol, data_dict, fsetup)

        # parse stereo
        if parse_stereo:

            data = get_sdf_field(molstr, 'stereoanalysis'.upper())
            if data is not None:
                for line in data:
                    tmp = line.split(' ')
                    if tmp[0] == 'CISTRANS':
                        # atoms enumeration in Chemaxon output starts from 0
                        id1 = int(tmp[1][1:-1]) + 1
                        id2 = int(tmp[2][:-1]) + 1
                        bond_stereo = tmp[-1].lower()
                        if bond_stereo == 'wiggly':
                            bond_stereo = 0
                        elif bond_stereo == 'e':
                            bond_stereo = 2
                        elif bond_stereo == 'z':
                            bond_stereo = 1
                        mol.SetDoubleBondConfig(id1, id2, bond_stereo)

            mol.SetCyclicDoubleBondsCis()

        return mol

    molstr = []
    i = 0
    with open(fname) as f:
        for line in f:
            if line.rstrip() != "$$$$":
                molstr.append(line.rstrip())
            else:
                i += 1
                yield _MolstrToMol(molstr, id_field_name, opt_diff, parse_stereo, i)
                molstr = []


def get_rdf_field(rxn, field_name):
    i = 0
    while rxn[i].find("$DTYPE") != 0:
        i += 1
    while rxn[i].split()[0] != "$DTYPE" and rxn[i].split()[1].lower() != field_name:
        i += 1
    i += 1
    field_data = ''
    # data can be on multiple lines
    while i < len(rxn) and rxn[i].find("$DTYPE") == -1:
        field_data += rxn[i].strip()
        i += 1
    return field_data.split(' ', 1)[1] if field_data else None


def parse_rxn_mols(rxn_lines):

    n_reactants = int(rxn_lines[4][0:3])
    n_products = int(rxn_lines[4][3:6])

    start_pos = 6
    pos = 6

    # read mols
    mols = []
    while n_reactants + n_products > len(mols):
        if rxn_lines[pos] != 'M  END':
            pos += 1
        else:
            mols.append(molstr_to_Mol(rxn_lines[start_pos:pos + 1]))
            start_pos = pos + 2
            pos = start_pos

    return mols, n_reactants, n_products


def set_mol_titles(mols, rx_id, n_reactants, n_products):
    for i in range(n_reactants):
        mols[i].title = rx_id + '_reactant_' + str(i + 1)
    for i in range(n_products):
        mols[n_reactants + i].title = rx_id + '_product_' + str(i + 1)


def create_rx_mix(mols, rx_id, n_reactants, n_products):
    mix = OrderedDict()
    mix[rx_id + '_reactants'] = {'names': [mols[i].title for i in range(n_reactants)],
                                 'ratios': [1] * n_reactants}
    mix[rx_id + '_products'] = {'names': [mols[n_reactants + i].title for i in range(n_products)],
                                'ratios': [1] * n_products}
    return mix


def process_rdf_reaction(rxn_lines, id_field_name, auto_name):

    mols, n_reactants, n_products = parse_rxn_mols(rxn_lines[1:])

    # get reaction id
    rx_id = None
    if id_field_name is not None:
        rx_id = get_rdf_field(rxn_lines, id_field_name)
    if rx_id is None:
        # from $RFMT $RIREG NUMBER
        tmp = rxn_lines[0].split()
        if len(tmp) == 3 and tmp[1] in ['$RIREG', '$REREG', '$MIREG', '$MEREG']:
                rx_id = tmp[2]
        else:
            rx_id = auto_name

    set_mol_titles(mols, rx_id, n_reactants, n_products)

    mix = create_rx_mix(mols, rx_id, n_reactants, n_products)

    # for now atomic properties cannot be considered for weighting of atoms in reactions
    # chemaxon cannot calculate atomic properties for reactions

    # read atom properties
    # data_dict = dict()
    # for diff in opt_diff:
    #     data = get_rdf_field(rxn_lines, diff)
    #     if data is not None:
    #         try:
    #             data_dict[diff] = [float(el.replace(",", ".")) if el != "" else None for el in data]
    #         except ValueError:
    #             data_dict[diff] = [el if el != "" else None for el in data]

    # add atom labels
    # if data_dict:
    #     n_atoms = 0
    #     for mol in mols:
    #         mol_data_dict = {k: v[n_atoms:len(mol.atoms)] for k, v in data_dict.items()}
    #         add_property_to_atoms(mol, mol_data_dict, fsetup)
    #         n_atoms += len(mol.atoms)

    return mols, mix


def process_rxn_reaction(rxn_lines, id_field_name, auto_name, opt_diff, fsetup):

    mols, n_reactants, n_products = parse_rxn_mols(rxn_lines)

    rx_id = auto_name
    if id_field_name is not None:
        data = get_sdf_field(rxn_lines, id_field_name)
        if data is not None and len(data) == 1:
            rx_id = data[0]

    # read properties from sdf fields
    data_dict = dict()
    for diff in opt_diff:
        data = get_sdf_field(rxn_lines, diff)
        if data is not None and len(data) == 1:
            s = data[0].split(';')
            try:
                data_dict[diff] = [float(el.replace(",", ".")) if el != "" else None for el in s]
            except ValueError:
                data_dict[diff] = [el if el != "" else None for el in s]

    # add labels
    start = 0
    for mol in mols:
        end = start + len(mol.atoms)
        data = {k: v[start:end] for k, v in data_dict.items()}
        add_property_to_atoms(mol, data, fsetup)
        start = end

    set_mol_titles(mols, rx_id, n_reactants, n_products)
    mix = create_rx_mix(mols, rx_id, n_reactants, n_products)

    return mols, mix


def ReadRDF(fname, id_field_name):
    # reaction ID will be searched in
    # 1) named field if specified (not None)
    # 2) $RFMT $RIREG id / $RFMT $REREG id if present
    # 3) generated automatically

    mols = OrderedDict()
    mix = OrderedDict()

    with open(fname) as rdf:

        # read file header
        rdf.readline()
        rdf.readline()

        rxn = []
        i = 1

        for line in rdf:
            if line.find("$RFMT") == 0:
                if rxn:
                    rx_mols, rx_mix = process_rdf_reaction(rxn, id_field_name, 'rx_autogen_id_' + str(i))
                    for mol in rx_mols:
                        mols[mol.title] = mol
                    mix.update(rx_mix)
                    i += 1
                rxn = [line.rstrip()]
            else:
                rxn.append(line.rstrip())

        rx_mols, rx_mix = process_rdf_reaction(rxn, id_field_name, 'rx_autogen_id_' + str(i))
        for mol in rx_mols:
            mols[mol.title] = mol
        mix.update(rx_mix)

    return mols, mix


def ReadRXN(fname, id_field_name, opt_diff, fsetup):

    mols = OrderedDict()
    mix = OrderedDict()

    with(open(fname)) as f:

        rxn = []
        i = 1

        for line in f:
            if line.find("$RXN") == 0:
                if rxn:
                    rx_mols, rx_mix = process_rxn_reaction(rxn, id_field_name, 'rx_autogen_id_' + str(i), opt_diff, fsetup)
                    for mol in rx_mols:
                        mols[mol.title] = mol
                    mix.update(rx_mix)
                    i += 1
                rxn = [line.rstrip()]
            else:
                rxn.append(line.rstrip())

        rx_mols, rx_mix = process_rxn_reaction(rxn, id_field_name, 'rx_autogen_id_' + str(i), opt_diff, fsetup)
        for mol in rx_mols:
            mols[mol.title] = mol
        mix.update(rx_mix)

    return mols, mix


#!/usr/bin/env python3
#-------------------------------------------------------------------------------
# Name:        sirms.py --in mols.sdf --out descriptors.txt
#              sirms.py -h for help
#
#
# Purpose:     calculate simplex descriptors for molecules in sdf file
#
# Author:      Pavel Polishchuk
#
# Created:     01.01.2013
# Copyright:   (c) Pavel Polishchuk 2013-2015
# Licence:     BSD 3-clause
# Python Version: 3.2
#-------------------------------------------------------------------------------

import argparse
import time
import copy
from itertools import combinations, chain, product, combinations_with_replacement
from collections import OrderedDict, defaultdict
from multiprocessing import Pool, cpu_count
from threading import Semaphore

from . import files
from . import sirms_name
from .sdf import ReadSDF, ReadRDF, ReadRXN
from .labels import SetLabelsInternal, builtin_types, GetSetupRanges, SetLabelsInternalToMol
from .ppgfunctions import *

mol_frag_sep = "###"


#===============================================================================
# Save simplexes

def sort_lists_by(*lists, key_list=0, desc=False):
    return zip(*sorted(zip(*lists), reverse=desc, key=lambda x: x[key_list]))


def SaveSimplexes(fname, sirms, output_format, ndigits=5):
    """
    Save calculated decriptors in the tab-delimited text file format
    Descriptors   descriptor_1  descriptor_2  ...  descriptor_n
    compound_1               5             2                  0
    compound_2               1             1                  4
    ...

    or in svm sparse format with two additional files with extensions colnames and rownames
    zero based indices
    0:12 1:23 4:1 ...
    """

    if output_format == 'txt':
        f_out = open(fname, 'wt')
        # get sorted unique simplex names
        sirms_names = sorted(list(set(list(chain.from_iterable([list(s.keys()) for s in sirms.values()])))))
        # compound_names = sorted([k for k in sirms.keys()])
        f_out.write("Compounds\t" + "\t".join(sirms_names) + "\n")
        s = "{:." + str(ndigits) + "f}"
        for n in sirms.keys():
            line = []
            for m in sirms_names:
                value = sirms[n].get(m, 0)
                # round floating point values, all others will be remained unchanged
                if type(value) is int or value.is_integer():
                    line.append(str(int(value)))
                elif isinstance(value, float):
                    line.append(s.format(value))
                else:
                    line.append(str(value))
            f_out.write(n + "\t" + "\t".join(map(str, line)) + "\n")
        f_out.close()
    elif output_format == 'svm':
        sirms_names = sorted(list(set(list(chain.from_iterable([list(s.keys()) for s in sirms.values()])))))
        colnames = sirms_names
        rownames = sirms.keys()
        sirms_names = dict(zip(sirms_names, range(len(sirms_names))))
        s = "{:." + str(ndigits) + "f}"
        missing_rows = []  # contains mols with all zero descriptors to omit in output
        with open(fname, 'wt') as f:
            for i, name in enumerate(sirms.keys()):
                if len(sirms[name]) == 0:
                    missing_rows.append(i)
                    continue
                # get new integer names of simplexes
                int_names = [sirms_names[v] for v in sirms[name].keys()]
                values = []
                for value in sirms[name].values():
                    # round floating point values, all others will be remained unchanged
                    if type(value) is int or value.is_integer():
                        values.append(str(int(value)))
                    elif isinstance(value, float):
                        values.append(s.format(value))
                    else:
                        values.append(str(value))
                int_names, values = sort_lists_by(int_names, values)
                f.write(' '.join([str(n) + ':' + v for n, v in zip(int_names, values)]) + '\n')
        open(os.path.splitext(fname)[0] + '.colnames', 'wt').write('\n'.join(colnames))
        open(os.path.splitext(fname)[0] + '.rownames', 'wt').write(
            '\n'.join([row for i, row in enumerate(rownames) if i not in missing_rows]))


#===============================================================================
# create reaction sirms
def concat_reaction_sirms(sirms):

    output = OrderedDict()

    while sirms:

        # take first element
        k, v = sirms.popitem(last=False)
        rx_id, role = k.rsplit('_', 1)
        # first letter from role is used as additional description
        new_sirms_names = [sirms_name.insert_reaction_info(s, role[0]) for s in sorted(v.keys())]
        s1 = {new_sirms_names[i]: v[sname] for i, sname in enumerate(sorted(v.keys()))}

        # take complimentary element (reactants or products of the same reaction)
        if role == 'reactants':
            k = rx_id + '_products'
            new_sirms_names = [sirms_name.insert_reaction_info(s, 'p') for s in sorted(sirms[k].keys())]
        elif role == 'products':
            k = rx_id + '_reactants'
            new_sirms_names = [sirms_name.insert_reaction_info(s, 'r') for s in sorted(sirms[k].keys())]
        else:
            print('Impossible error')
        v = sirms[k]
        s2 = {new_sirms_names[i]: v[sname] for i, sname in enumerate(sorted(sirms[k].keys()))}
        del (sirms[k])

        # combine them
        s1.update(s2)
        output[rx_id] = s1

    return output


#===============================================================================
# SIRMS

def prep_input(in_fname, id_field_name, opt_diff, opt_diff_sdf, setup_path, min_num_atoms, max_num_atoms,
               min_num_components, max_num_components, noH, verbose, per_atom_fragments, frags, semaphore):

    ranges = GetSetupRanges(opt_diff, setup_path)

    for mol in ReadSDF(in_fname, id_field_name, opt_diff_sdf, setup_path):

        semaphore.acquire()

        # get part of a frag dict for the current mol
        mol_frag = OrderedDict()
        if frags is not None and mol.title in frags:
            mol_frag.update(frags[mol.title])
        if per_atom_fragments:
            mol_frag.update(GetPerAtomMolFragments(mol, noH))

        # set internal labels (labels from SDF automatically loaded)
        SetLabelsInternalToMol(mol, opt_diff, ranges)

        yield mol, opt_diff, min_num_atoms, max_num_atoms, min_num_components, max_num_components, noH, verbose, mol_frag


def prep_input_mix(mol_list, opt_diff, min_num_atoms, max_num_atoms, min_num_components, max_num_components,
                   opt_noH, opt_verbose):
    for mol in mol_list:
        yield mol, opt_diff, min_num_atoms, max_num_atoms, min_num_components, max_num_components, opt_noH, opt_verbose


def MapCalcMolSingleSirms(args):
    return CalcMolSingleSirms(*args)


def CalcMolSingleSirms(mol, diff_list, min_num_atoms, max_num_atoms, min_num_components,
                       max_num_components, noH, verbose, frags_dict=None):
    """
    INPUT:
        mol: Mol3 object;
        sirms_dict: loaded precomputed dictionary;
        diff_list: list of labeling to compute;
    OUTPUT:
        returns dict of dicts
        if frags are specified then it returns descriptors for fragment-depleted molecule
        { "mol_name":               { "descriptors_1": 0, "descriptor_2": 5, ... },
          "mol_name#frag_name_1":   { "descriptors_1": 0, "descriptor_2": 3, ... },
          ...
        }

        in case of for_mix=True
        returns additional dict with sirms names and atom counts
    """

    if verbose:
        cur_time = time.time()

    # prepare output dict
    d = OrderedDict()
    d[mol.title] = defaultdict(int)
    if frags_dict:
        local_frags_dict = OrderedDict((mol.title + mol_frag_sep + k, set(v)) for k, v in frags_dict.items())
        for full_frag_name in local_frags_dict:
            d[full_frag_name] = defaultdict(int)
    else:
        local_frags_dict = None

    for a in mol.GetAtomsCombinations(min_num_components=min_num_components, max_num_components=max_num_components,
                                      min_num_atoms=min_num_atoms, max_num_atoms=max_num_atoms, noH=noH):

        for s_diff in diff_list:
            labels = [mol.atoms[a_id]['property'][s_diff]['label'] for a_id in a]
            for labels_set in product(*labels):
                s_name = sirms_name.create_full_name(prod_react='', single_mix='S', num_prob='n',
                                                     atom_count=len(a), atom_labeling=s_diff,
                                                     smiles=mol.get_name(a, labels_set))
                d[mol.title][s_name] += 1
                if local_frags_dict:
                    # if there is no common atoms in simplex and fragment
                    set_a = set(a)
                    for frag_name in local_frags_dict:
                        if set_a.isdisjoint(local_frags_dict[frag_name]):
                            d[frag_name][s_name] += 1

    if verbose:
        print(mol.title, (str(round(time.time() - cur_time, 1)) + ' s').rjust(78 - len(mol.title)))

    d = OrderedDict((k, dict(v)) for k, v in d.items())

    return d


#===============================================================================

def CalcSingleSirms(mol_list, opt_diff, min_num_atoms, max_num_atoms, min_num_components,
                    max_num_components, opt_noH, opt_verbose, frags):

    if opt_verbose:
        print(' Single compounds calculation '.center(79, '='))

    sirms = OrderedDict()
    for mol in mol_list:
        res = CalcMolSingleSirms(mol, opt_diff, min_num_atoms, max_num_atoms, min_num_components,
                                 max_num_components, opt_noH, opt_verbose, frags)
        sirms.update(res)

    return sirms


#===============================================================================

def CalcMixSirms(single_sirms, mix, atom_labeling, min_num_atoms=4, max_num_atoms=4,
                 min_num_mix_components=2, max_num_mix_components=2, verbose=False, ordered=False,
                 self_assembly_mix=False):

    def mult(values):
        p = 1
        for item in values:
            p *= item
        return p

    def gen_mix_sirms_name(sirms_names, ordered, ids=None):
        if not ordered:
            tmp = [sirms_name.get_smile(item) for item in sirms_names]
            tmp = [item.split('.') for item in tmp]
            # flatten list and sort to obtain canonical order
            tmp = sorted([item for sublist in tmp for item in sublist])
        else:
            # add component id before smile
            tmp = sorted([str(id + 1) + '_' + sirms_name.get_smile(item) for id, item in zip(ids, sirms_names)])
        # get atom count
        atom_count = sum(sirms_name.get_atomcount(item) for item in sirms_names)
        # get atomic property of descriptors (they all must have the same)
        return sirms_name.create_full_name(prod_react='', single_mix='M', num_prob='n', atom_count=atom_count,
                                           atom_labeling=sirms_name.get_atom_labeling(sirms_names[0]),
                                           smiles='.'.join(tmp))

    def select_sirms_by_labeling_one_component(sirms_names, labeling_name):
        substr = '|' + labeling_name + '|'
        return [item for item in sirms_names if item.find(substr) > -1 and item.split('|')[-1].find('.') == -1]

    if verbose:
        print(' Mixtures calculation '.center(79, '='))

    d = OrderedDict()

    for mix_name, mix_data in mix.items():

        if verbose:
            cur_time = time.time()

        if self_assembly_mix:
            iterator = combinations_with_replacement
        else:
            iterator = combinations

        d_mix = dict()

        # multiply descriptors on compounds ratios
        d_tmp = dict()
        for i, mol_name in enumerate(mix_data['names']):
            d_tmp[mol_name] = copy.deepcopy(single_sirms[mol_name])
            for name, value in d_tmp[mol_name].items():
                d_tmp[mol_name][name] = value * mix_data['ratios'][i]

        for n in range(min_num_mix_components, max_num_mix_components + 1):
            for ids in iterator(range(len(mix_data['names'])), n):   # ids - mol id in mixture
                comb = [mix_data['names'][i] for i in ids]
                for labeling in atom_labeling:
                    for p in product(*[select_sirms_by_labeling_one_component(d_tmp[mol_name].keys(), labeling) for mol_name in comb]):  # p - combination of sirms names from molecules
                        s = sum(sirms_name.get_atomcount(item) for item in p)
                        if min_num_atoms <= s <= max_num_atoms:
                            mix_sirs_name = gen_mix_sirms_name(p, ordered, ids)
                            d_mix[mix_sirs_name] = d_mix.get(mix_sirs_name, 0) + mult([d_tmp[mol_name].get(p[i], 0) for i, mol_name in enumerate(comb)])

        # add single sirms and filter them with given number of atoms
        sirms_names = set(list(chain.from_iterable(list(s.keys()) for s in single_sirms.values())))
        for name in sirms_names:
            if min_num_atoms <= sirms_name.get_atomcount(name) <= max_num_atoms:
                d_mix[name] = sum(single_sirms[mol_name].get(name, 0) for mol_name in mix_data['names'])

        d[mix_name] = d_mix

        if verbose:
            print(mix_name, (str(round(time.time() - cur_time, 1)) + ' s').rjust(78 - len(mix_name)))

    return d


#===============================================================================

def GenQuasiMix(mol_names):
    d = OrderedDict()
    for n in mol_names:
        d[n] = {'names': [n, n], 'ratios': [0.5, 0.5]}
    return d


#===============================================================================

def GetPerAtomMolFragments(mol, noH):
    d = OrderedDict()
    counter = 0
    for i, a in enumerate(mol.atoms):
        if noH and mol.atoms[a]["label"] == 'H':
            continue
        d["%i#%i" % (i+1, counter)] = [i + 1]
        counter += 1
    return d

#===============================================================================

def CalcReactionDiffSirms(sirms):

    def calc_diff(d):
        output = dict()
        checked_names = set()
        for name, value in d.items():
            r, common_part = sirms_name.split_by_reaction_part(name)
            if common_part not in checked_names:
                checked_names.add(common_part)
                if r == 'p':
                    res = d[name] - d.get(sirms_name.join_reaction_part('r', common_part), 0)
                elif r == 'r':
                    res = d.get(sirms_name.join_reaction_part('p', common_part), 0) - d[name]
                else:
                    print("Incorrect reaction_part label %s, can be only r, p or pr." % r)
                    continue
                if res != 0:
                    output[sirms_name.join_reaction_part('pr', common_part)] = res
        return output


    for mol_name, data in sirms.items():
        sirms[mol_name] = calc_diff(sirms[mol_name])
    return sirms


#===============================================================================

def CalcProbSirms(sirms, type):
    if type in ['prob', 'both']:
        for mol_name in sirms.keys():
            d = dict()
            s_mix = sum(v for k, v in sirms[mol_name].items() if sirms_name.get_mix_single(k) == 'M')
            s_single = sum(v for k, v in sirms[mol_name].items() if sirms_name.get_mix_single(k) == 'S')
            for k, v in sirms[mol_name].items():
                if sirms_name.get_mix_single(k) == 'M':
                    d[sirms_name.invert_num_prob_type(k)] = v/s_mix if s_mix != 0 else 0
                else:
                    d[sirms_name.invert_num_prob_type(k)] = v/s_single if s_single != 0 else 0
            if type == 'both':
                sirms[mol_name].update(d)
            elif type == 'prob':
                sirms[mol_name] = d
    return sirms


#===============================================================================
# Main cycle

def main_params(in_fname, out_fname, opt_diff, min_num_atoms, max_num_atoms, min_num_components,
                max_num_components, min_num_mix_components, max_num_mix_components, mix_fname,
                descriptors_transformation, mix_type, opt_mix_ordered, opt_verbose, opt_noH,
                frag_fname, per_atom_fragments, self_association_mix, reaction_diff, quasimix, id_field_name,
                output_format, ncores):

    # define which property will be loaded from external file or from sdf-file
    opt_diff_builtin = [v for v in opt_diff if v in builtin_types]
    opt_diff_sdf = [v for v in opt_diff if v not in builtin_types]

    # load sdf, rdf or rxn file depending on its extension
    input_file_extension = in_fname.strip().split(".")[-1].lower()
    setup_path = os.path.join(GetWorkDir(in_fname), "setup.txt")

    # check all properties are present in setup
    if opt_diff_sdf:
        not_avail = set(opt_diff_sdf).difference(files.GetAtomPropertyFromSetup(setup_path))
        if not_avail:
            for v in not_avail:
                print("WARNING. Chosen atomic property values (%s) is absent in setup.txt file. "
                      "Therefore its values will used as categorical variable ('as is') for atom labeling." % v)

    # init pool of workers for calculation of single compounds
    ncores = min(cpu_count(), max(ncores, 1))
    p = Pool(ncores)
    chunksize = 5
    semaphore = Semaphore(ncores * chunksize)

    if mix_fname is None and not quasimix and input_file_extension == 'sdf':

        saver = None
        sirms = None
        if output_format == "svm":
            saver = files.SvmSaver(out_fname)
        if output_format == "txt":
            sirms = OrderedDict()
        frags = files.LoadFragments(frag_fname)

        try:
            for result in p.imap(MapCalcMolSingleSirms,
                                 prep_input(in_fname, id_field_name, opt_diff, opt_diff_sdf, setup_path, min_num_atoms,
                                            max_num_atoms, min_num_components, max_num_components, opt_noH,
                                            opt_verbose, per_atom_fragments, frags, semaphore),
                                 chunksize=chunksize):
                if output_format == "txt":
                    sirms.update(result)
                    semaphore.release()
                if output_format == "svm":
                    for mol_name, descr_dict in result.items():
                        saver.save_mol_descriptors(mol_name, descr_dict)
                        semaphore.release()

        finally:
            p.close()

        if output_format == "txt":
            SaveSimplexes(out_fname, sirms, output_format)

    else:

        # read input
        if input_file_extension == 'sdf':
            mols = OrderedDict()
            for m in ReadSDF(in_fname, id_field_name, opt_diff_sdf, setup_path):
                mols[m.title] = m
        elif input_file_extension == 'rdf':
            mols, mix = ReadRDF(in_fname, id_field_name)
        elif input_file_extension == 'rxn':
            mols, mix = ReadRXN(in_fname, id_field_name, opt_diff_sdf, setup_path)
        else:
            print("Input file extension should be SDF, RDF or RXN. Current file has %s. Please check it." %
                  input_file_extension.upper())
            return None

        # set labels of built-in types
        SetLabelsInternal(mols, opt_diff_builtin, setup_path)

        # create mix data for sdf (for rdf/rxn mix is created during file loading)
        if input_file_extension == 'sdf':
            if quasimix:
                mix = GenQuasiMix(list(mols.keys()))
            elif mix_fname is not None:
                mix = files.LoadMixturesTxt(mix_fname, mix_type)
            else:
                print("Strange error occurred during mix preparation")
                exit()

        mols_used = set(chain.from_iterable([m['names'] for m in mix.values()]))

        sirms = OrderedDict()
        try:
            for result in p.imap(MapCalcMolSingleSirms,
                                 prep_input_mix([mols[mol_name] for mol_name in mols_used], opt_diff, 1, max_num_atoms,
                                                min_num_components, max_num_components, opt_noH, opt_verbose),
                                 chunksize=chunksize):
                sirms.update(result)
        finally:
            p.close()

        # min_num_atoms set to 1 to be able to generate mixtures
        # sirms = CalcSingleSirms([mols[mol_name] for mol_name in mols_used], opt_diff, 1,
        #                         max_num_atoms, min_num_components, max_num_components, opt_noH,
        #                         opt_verbose, None)

        sirms = CalcMixSirms(single_sirms=sirms,
                             mix=mix,
                             atom_labeling=opt_diff,
                             min_num_atoms=min_num_atoms,
                             max_num_atoms=max_num_atoms,
                             min_num_mix_components=min_num_mix_components,
                             max_num_mix_components=max_num_mix_components,
                             verbose=opt_verbose,
                             ordered=opt_mix_ordered,
                             self_assembly_mix=self_association_mix)

        # filter single sirms with number of atoms lower than min_num_atoms

        if input_file_extension in ['rdf', 'rxn']:
            sirms = concat_reaction_sirms(sirms)

        if descriptors_transformation in ['prob', 'both']:
            sirms = CalcProbSirms(sirms, descriptors_transformation)

        if reaction_diff:
            sirms = CalcReactionDiffSirms(sirms)

        SaveSimplexes(out_fname, sirms, output_format)


def main():
    parser = argparse.ArgumentParser(description='Calculate simplex descriptors for single molecules, quasi-mixtures, '
                                                 'mixtures and reactions.')
    parser.add_argument('-i', '--in', metavar='input.sdf', required=False, default=None,
                        help='input file (allowed formats: sdf, rdf, rxn) with standardized structures, '
                             'molecules or reactions should have titles.')
    parser.add_argument('-o', '--out', metavar='output.txt', required=False, default=None,
                        help='output file with calculated descriptors. Can be in text or sparse svm format.')
    parser.add_argument('-b', '--output_format', metavar='output_format', default='txt',
                        help='format of output file with calculated descriptors (txt|svm). '
                             'Txt is ordinary tab-separated text file. Svm is sparse format, two additional files will '
                             'be saved with extensions .colnames and .rownames. Default: txt.')
    parser.add_argument('-a', '--atoms_labeling', metavar='', default=['elm'], nargs='*',
                        help='list of atom labeling schemes separated by space. Built-in scheme is element (elm) and '
                             'topology (none). '
                             'To include other schemes user should specify the name of the corresponding property '
                             'value identical to the name of SDF field, which contains calculated atomic properties. '
                             'Fields names are case-sensitive. For RDF/RXN input files only built-in types can be '
                             'used. Default: elm.')
    parser.add_argument('--min_atoms', metavar='', default=4,
                        help='The minimal number of atoms in the fragment. Default: 4')
    parser.add_argument('--max_atoms', metavar='', default=4,
                        help='The maximal number of atoms in the fragment. Default: 4')
    parser.add_argument('--min_components', metavar='', default=1,
                        help='The minimal number of disconnected groups of atoms in the fragment. '
                             'Default: 1 (mean fully connected fragments).')
    parser.add_argument('--max_components', metavar='', default=2,
                        help='The maximal number of disconnected groups of atoms in the fragment. '
                             'Default: 2.')
    parser.add_argument('-q', '--quasi_mix', action='store_true', default=False,
                        help='calculate quasi-mixture descriptors for pure compounds.')
    parser.add_argument('-m', '--mixtures', metavar='mixtures.txt', default=None,
                        help='text file containing list of mixtures of components and their ratios. Names of '
                             'components should be the same as in input.sdf file.')
    parser.add_argument('--min_mix_components', metavar='', default=2,
                        help='The minimal number of molecules which contribute to mixture fragments. '
                             'Default: 2 (and cannot be less).')
    parser.add_argument('--max_mix_components', metavar='', default=2,
                        help='The maximal number of molecules which contribute to mixture fragments. '
                             'Default: 2 (take into account only binary interactions in a mixture).')
    parser.add_argument('--descriptors_transformation', metavar='num|prob|both', default='num',
                        help='num: numbers of fragments (for single compounds) or number of fragments combinations '
                             'weighted by their molar ratios (for mixtures). '
                             'prob: final value of descriptors are divided on sum of all descriptors to '
                             'describe probability of each descriptor. Descriptors for single compounds and mixtures '
                             'are weighted separately. both: will generate both types of descriptors. Default: num.')
    parser.add_argument('--mix_type', metavar='abs|rel', default='abs',
                        help='abs: means that amount of components given in a mixture file will be considered as is. '
                             'rel: means that amount of components given in a mixture file will be taken as relative '
                             'amount and will be scaled to the sum of 1. Default: abs.')
    parser.add_argument('-r', '--mix_ordered', action='store_true', default=False,
                        help='if set this flag the mixtures will be considered ordered, otherwise as unordered. '
                             'In ordered mixtures role of each component is known and position of a component in a '
                             'mixture description file will be taken into account. In unordered mixtures all '
                             'components are equitable and their roles don''t depend on their positions in mixture '
                             'description. Used only in combination with -m key.')
    parser.add_argument('--mix_self_association', action='store_true', default=False,
                        help='calculates mixture descriptors between components with themselves in order to take into '
                             'account self-interaction of components. Default: false.')
    parser.add_argument('--reaction_diff', action='store_true', default=False,
                        help='if set this flag difference between product and reactant descriptors will be calculated. '
                             'By default these two feature vectors are concatenated.')
    parser.add_argument('-v', '--verbose', action='store_true', default=False,
                        help='if set this flag progress will be printed out (may cause decrease in speed).')
    parser.add_argument('-x', '--noH', action='store_true', default=False,
                        help='if set this flag hydrogen atoms will be excluded from the simplexes calculation.')
    parser.add_argument('-f', '--fragments', metavar='fragments.txt', default=None,
                        help='text file containing list of names of single compounds, fragment names and atom '
                             'indexes of fragment to remove (all values are tab-separated).')
    parser.add_argument('--per_atom_fragments', action='store_true', default=False,
                        help='if set this flag input fragments will be omitted and single atoms will be considered '
                             'as fragments.')
    parser.add_argument('-w', '--id_field_name', metavar='field_name', default=None,
                        help='field name of unique ID for compounds (sdf) or reactions (rdf/rxn). '
                             'If omitted for sdf molecule titles will be used or auto-generated names; '
                             'for rdf $RIREG/$REREG/$MIREG/$MEREG field if it is not empty or auto-generated names')
    parser.add_argument('--version', action='store_true', default=False,
                        help='print the version of the program and exit.')
    parser.add_argument('-c', '--ncores', metavar='number_of_cores', default=1,
                        help='number of cores used for calculation of descriptors for single molecules only.')

    args = vars(parser.parse_args())
    opt_mix_ordered = None
    for o, v in args.items():
        if o == "in": in_fname = v
        if o == "out": out_fname = v
        if o == "atoms_labeling": opt_diff = v
        if o == "mixtures": mix_fname = v
        if o == "mix_ordered": opt_mix_ordered = v
        if o == "verbose": opt_verbose = v
        if o == "noH": opt_noH = v
        if o == "fragments": frag_fname = v
        if o == "per_atom_fragments": per_atom_fragments = v
        if o == "quasi_mix": quasimix = v
        if o == "id_field_name": id_field_name = v
        if o == "output_format": output_format = v
        if o == "min_atoms": min_num_atoms = int(v)
        if o == "max_atoms": max_num_atoms = int(v)
        if o == "min_components": min_num_components = int(v)
        if o == "max_components": max_num_components = int(v)
        if o == "min_mix_components": min_num_mix_components = int(v)
        if o == "max_mix_components": max_num_mix_components = int(v)
        if o == "mix_self_association": self_association_mix = v
        if o == "descriptors_transformation": descriptors_transformation = v
        if o == "mix_type": mix_type = v
        if o == "reaction_diff": reaction_diff = v
        if o == "version": version = v
        if o == "ncores": ncores = int(v)

    if version:
        print('version 1.1.1')
        exit()

    if not version:
        if in_fname is None:
            print("Please specify the input file.")
            exit()
        if out_fname is None:
            print("Please specify the output file.")
            exit()

    if quasimix:
        opt_mix_ordered = False
        mix_fname = None
    if in_fname.split('.')[-1] in ['rdf', 'rxn']:
        opt_mix_ordered = False
        mix_fname = None
    if output_format not in ['svm', 'txt']:
        print("INPUT ERROR: wrong output format specified - %s. Only txt or svm are allowed." % output_format)
        exit()
    if min_num_atoms > max_num_atoms:
        print("INPUT ERROR: min_num_atoms should not be greater than max_num_atoms.")
        exit()
    if min_num_components > max_num_components:
        print("INPUT ERROR: min_num_components should not be greater than max_num_components.")
        exit()
    if min_num_mix_components > max_num_mix_components or min_num_mix_components < 2:
        print("INPUT ERROR: min_num_mix_components should not be greater than max_num_mix_components and "
              "minimal value of min_num_mix_components should be 2 or greater.")
        exit()
    if quasimix or mix_fname is not None or in_fname.split('.')[-1] in ['rdf', 'rxn']:
        if min_num_mix_components > max_num_components:
            print("INPUT ERROR: minimal number of mixture components (min_num_mix_components) cannot be greater than "
                  "maximal number of components in a fragment descriptor (min_num_components).")
            exit()
    if mix_type not in ['abs', 'rel']:
        print("INPUT ERROR: mixture type (mix_type) can be only abs or rel.")
        exit()
    if descriptors_transformation not in ['num', 'prob', 'both']:
        print("INPUT ERROR: type of mixture descriptors (types_mix_descriptors) can be only num, prob or both.")
        exit()
    if in_fname.split('.')[-1] not in ['rdf', 'rxn'] and reaction_diff:
        print("The option reaction diff can be used only with RDF or RXN input files.")
        exit()

    main_params(in_fname=in_fname,
                out_fname=out_fname,
                opt_diff=opt_diff,
                min_num_atoms=min_num_atoms,
                max_num_atoms=max_num_atoms,
                min_num_components=min_num_components,
                max_num_components=max_num_components,
                min_num_mix_components=min_num_mix_components,
                max_num_mix_components=max_num_mix_components,
                mix_fname=mix_fname,
                descriptors_transformation=descriptors_transformation,
                mix_type=mix_type,
                opt_mix_ordered=opt_mix_ordered,
                opt_verbose=opt_verbose,
                opt_noH=opt_noH,
                frag_fname=frag_fname,
                per_atom_fragments=per_atom_fragments,
                self_association_mix=self_association_mix,
                reaction_diff=reaction_diff,
                quasimix=quasimix,
                id_field_name=id_field_name,
                output_format=output_format,
                ncores=ncores)


if __name__ == '__main__':
    main()



import setuptools
from os import path
import sirms

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setuptools.setup(
    name="sirms",
    version=sirms.__version__,
    author="Pavel Polishchuk",
    author_email="pavel_polishchuk@ukr.net",
    description="SiRMS: simplex representation of molecular structure",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/DrrDom/sirms",
    packages=['sirms'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Chemistry"
    ],
    python_requires='>=3.6',
    extras_require={
        'rdkit': ['rdkit>=2017.09'],
    },
    entry_points={'console_scripts':
                  ['sirms = sirms.sirms:main',
                   'calc_atomic_properties_chemaxon = sirms.utilities.calc_atomic_properties_chemaxon:main']}
)



#-------------------------------------------------------------------------------
# Name:        files
# Purpose:     working with general purpose files
#
# Author:      Pavel Polishchuk
#
# Created:     06.06.2013
# Copyright:   (c) Pavel Polishchuk 2013-2015
# Licence:     BSD 3-clause
#-------------------------------------------------------------------------------

import os
from string import ascii_uppercase
from .ppgfunctions import GetFileNameNoExt, GetWorkDir
from collections import OrderedDict


class SvmSaver:

    def __init__(self, file_name):
        self.__fname = file_name
        self.__varnames_fname = os.path.splitext(file_name)[0] + '.colnames'
        self.__molnames_fname = os.path.splitext(file_name)[0] + '.rownames'
        self.__varnames = []
        # self.__molnames = []
        if os.path.isfile(self.__fname):
            os.remove(self.__fname)
        if os.path.isfile(self.__molnames_fname):
            os.remove(self.__molnames_fname)
        if os.path.isfile(self.__varnames_fname):
            os.remove(self.__varnames_fname)

    def __convert_value(self, value):
        if type(value) is int or value.is_integer():
            res = str(int(value))
        elif isinstance(value, float):
            res = "{:.5f}".format(value)
        else:
            res = str(value)
        return res

    def save_mol_descriptors(self, mol_name, mol_descr_dict):

        values = []

        for i, varname in enumerate(self.__varnames):
            if varname in mol_descr_dict:
                value = self.__convert_value(mol_descr_dict[varname])
                values.append((i, value))

        new_varnames = list(set(mol_descr_dict) - set(self.__varnames))
        for i, varname in enumerate(new_varnames):
            value = self.__convert_value(mol_descr_dict[varname])
            values.append((len(self.__varnames) + i, value))

        if values:  # values can be empty if all descriptors are zero

            self.__varnames.extend(new_varnames)

            with open(self.__molnames_fname, 'at') as f:
                f.write(mol_name + '\n')

            if new_varnames:
                with open(self.__varnames_fname, 'at') as f:
                    f.write('\n'.join(new_varnames) + '\n')

            with open(self.__fname, 'at') as f:
                values = sorted(values)
                values = ('%i:%s' % (i, v) for i, v in values)
                f.write(' '.join(values) + '\n')


def GetAtomPropertyFromSetup(setup_fname):
    with open(setup_fname) as f:
        output = [line.strip().split('=')[0] for line in f]
    return output


# def NotExistedPropertyFiles(opt_diff, input_fname):
#     """
#     Checks property files existence and returns names of not existed property files,
#     these properties will be loaded from the initial sdf file
#     """
#     output = []
#     for s_diff in opt_diff:
#         if s_diff not in ["elm", "none", "uffd", "uffe"]:   # built-in types are ignored
#             if not os.path.isfile(os.path.join(GetWorkDir(input_fname),
#                                                GetFileNameNoExt(input_fname) + '.' + s_diff)):
#                 output.append(s_diff)
#     return (output)


def ReadPropertyRange(file_setup_name, prop_name):
    res = None
    try:
        f = open(file_setup_name)
        for line in f:
            if prop_name == line.split('=')[0]:
                res = list(map(float, line.strip().split('=')[1].split('<')))
                break
    except FileNotFoundError:
        print('File setup.txt cannot be found in the folder containing input sdf file.')
        exit()
    else:
        f.close()
    return res


def RangedLetter(value, prop_range, none_label="NA"):
    if value is None:
        return (none_label)
    if value <= prop_range[0]:
        return ('A')
    for i in range(1, len(prop_range)):
        if prop_range[i - 1] < value <= prop_range[i]:
            return (ascii_uppercase[i])
        if i == len(prop_range) - 1 and value > prop_range[-1]:
            return (ascii_uppercase[i + 1])
    print('Property label was not assigned.')
    return (None)


def LoadRangedProperty(mols, setup_dir, prop_fname):

    def ReadPropertyFile(fname):
        f = open(fname)
        sep = '---'
        d = {}
        f.readline()  # header
        # read each mol as list of lists where inner list is ['atom_id', 'value']
        while True:
            line = f.readline()
            if not line: break
            if line.strip() == sep:
                title = f.readline().strip()
                d[title] = list()
                continue
            d[title].append(line.strip().split())
        f.close()
        # sort values for each mol by atom_id and keep values only
        for k in d.keys():
            d[k].sort(key=lambda x: int(x[0]))
            # if value is not numeric remain it as is (as string)
            try:
                d[k] = [float(i[1]) for i in d[k]]
            except ValueError:
                d[k] = [i[1] for i in d[k]]
        return (d)

    prop_name = os.path.splitext(os.path.basename(prop_fname))[1][1:]
    prop_range = ReadPropertyRange(os.path.join(setup_dir, 'setup.txt'), prop_name)
    # read property file
    d = ReadPropertyFile(os.path.join(setup_dir, prop_fname))
    # assign property values to atoms
    for m in mols.values():
        values = d[m.title]
        for i, a in enumerate(sorted(m.atoms.keys())):
            # if label is string use it as label, otherwise use ranges
            label = values[i] if type(values[i]) is str else RangedLetter(values[i], prop_range)
            m.atoms[a]['property'][prop_name] = {'value': values[i], 'label': label}


def LoadMixturesTxt(fname, mix_type):
    """
    Load mixture composition
    INPUT: tab-delimited text file
    mol_name_1  mol_name_2  first_component_ratio  second_component_ratio
    mol_name_1  mol_name_3  first_component_ratio  second_component_ratio
    ...
    NOTE: number of components can be any
    OUTPUT:
    { "mol_name_1_(ratio_1)+mol_name_2_(ratio_2)":
        {'names': ["mol_name_1", "mol_name_2"], 'ratios': [ratio_1, ratio_2]},
      "mol_name_1_(ratio_1)+mol_name_3_(ratio_2)":
        {'names': ["mol_name_1", "mol_name_3"], 'ratios': [ratio_1, ratio_2]},
      ...
    }
    OUTPUT EXAMPLE:
    { 'P-Ser-Ser_(1.0)+E-37_(1.0)': {'names': ['P-Ser-Ser', 'E-37'], 'ratios': [1.0, 1.0]},
      'K-Arg-Ser_(1.0)+MS-7_(1.0)': {'names': ['K-Arg-Ser', 'MS-7'], 'ratios': [1.0, 1.0]},
      'K-Ala-Glu_(1.0)+E-21_(1.0)': {'names': ['K-Ala-Glu', 'E-21'], 'ratios': [1.0, 1.0]} }
    """
    d = OrderedDict()

    if mix_type == 'abs':
        abs_ratio = True
    else:
        abs_ratio = False

    f = open(fname, 'rt')

    while True:
        line = f.readline()
        if not line: break
        tmp = line.strip().split('\t')
        mix_num = len(tmp) // 2
        mol_names = tmp[:mix_num]
        mol_ratios = list(map(float, tmp[(len(tmp) // 2):]))
        if not abs_ratio:
            mol_ratios = [v / sum(mol_ratios) for v in mol_ratios]
        mol_ratios = [round(v, 4) for v in mol_ratios]
        mix_name = '+'.join([mol_names[i] + '_(' + str(mol_ratios[i]) + ')' for i in range(mix_num)])
        d[mix_name] = {'names': mol_names, 'ratios': mol_ratios}
    return d


def LoadFragments(fname):
    """
    INPUT
    Input file format - tab delimited text file, each line contains
    mol_name  fragment_name  atom_index_1  atom_index_2 ... atom_index_n
    where atom ids of fragment atoms in the molecule are listed
    OUTPUT
    Output dict of molecules which are dict of fragments with corresponding atom ids
    { "mol_name_1":
        {
          "frag_name_1#0": [atom_id1, atom_id2, ...],
          "frag_name_1#1": [atom_id1, atom_id2, ...],
          "frag_name_2#0": [atom_id1, atom_id2, ...]
        }
      ...
    }
    """
    if fname is None:
        return (None)
    d = OrderedDict()
    f = open(fname, 'rt')
    for line in f:
        tmp = line.strip().split("\t")
        if tmp[0] not in d.keys():
            d[tmp[0]] = OrderedDict()
        d[tmp[0]][tmp[1] + "#" + str(len(d[tmp[0]]))] = list(map(int, tmp[2:]))
    f.close()
    return (d)


#-------------------------------------------------------------------------------
# Name:        ppgfunctions
# Purpose:     general purpose functions
#
# Author:      Pavel Polishchuk
#
# Created:     11.01.2013
# Copyright:   (c) Pavel Polishchuk 2013-2015
# Licence:     BSD 3-clause
#-------------------------------------------------------------------------------

import os


def SortTwoLists(primary, secondary):
    # sort two lists by order of elements of the primary list
    paired_sorted = sorted(zip(primary, secondary), key=lambda x: x[0])
    return (map(list, zip(*paired_sorted)))  # two lists


def GetFileNameNoExt(path):
    return (os.path.splitext(os.path.basename(path)))[0]


def GetFileExt(path):
    return (os.path.splitext(os.path.basename(path)))[1]


def GetWorkDir(filename):
    return (os.path.abspath(os.path.dirname(filename)))
